@@ -79,8 +79,8 @@ GString::~GString() {
 }
 
 GString *GString::clear() {
-  resize(0);
   s[length = 0] = '\0';
+  resize(0);
   return this;
 }
 
@@ -109,6 +109,14 @@ GString *GString::append(char *str) {
   return this;
 }
 
+GString *GString::append(char *str, int length1) {
+  resize(length + length1);
+  memcpy(s + length, str, length1);
+  length += length1;
+  s[length1] = '\0';
+  return this;
+}
+
 GString *GString::insert(int i, char c) {
   int j;
 
