@@ -18,7 +18,6 @@
 class Dict;
 class XRef;
 class OutputDev;
-class PSOutput;
 
 //------------------------------------------------------------------------
 // PageAttrs
@@ -32,17 +31,23 @@ public:
   // <attrs> is NULL, uses defaults.
   PageAttrs(PageAttrs *attrs, Dict *dict);
 
+  // Destructor.
+  ~PageAttrs();
+
   // Accessors.
   int getX1() { return x1; }
   int getY1() { return y1; }
   int getX2() { return x2; }
   int getY2() { return y2; }
   int getRotate() { return rotate; }
+  Dict *getResourceDict()
+    { return resources.isDict() ? resources.getDict() : (Dict *)NULL; }
 
 private:
 
   int x1, y1, x2, y2;
   int rotate;
+  Object resources;
 };
 
 //------------------------------------------------------------------------
@@ -61,9 +66,10 @@ public:
   // Is page valid?
   GBool isOk() { return ok; }
 
-  // Get page size.
+  // Get page parameters.
   int getWidth() { return attrs->getX2() - attrs->getX1(); }
   int getHeight() { return attrs->getY2() - attrs->getY1(); }
+  int getRotate() { return attrs->getRotate(); }
 
   // Get font dictionary.
   Dict *getFontDict()
@@ -75,9 +81,6 @@ public:
   // Display a page.
   void display(OutputDev *out, int dpi, int rotate);
 
-  // Generate PostScript for a page.
-  void genPostScript(PSOutput *psOut, int dpi, int rotate);
-
 private:
 
   int num;			// page number
